"""ME-GPT: A library for creating AI-powered personality clones using vector databases

This library allows users to create digital "clones" of themselves by storing memories
and personality traits in a local vector database, which can then be used to power
conversations with an AI that mimics their personality.

For more information, please visit:
https://github.com/djm93dev/me_gpt
"""

__version__ = "0.1.6"
__author__ = "Daniel McDonald"
__author_email__ = "daniel@danielmcdonald.dev"
__license__ = "MIT"
__copyright__ = "Copyright 2024 Daniel McDonald"

# Core functionality
from me_gpt.ai import (
    initial_questions,
    more_questions,
    prompt_memory,
    chat_with_clone,
    chat,
    clear_chat_memory,
)

# Database and similarity functions
from me_gpt.db import DB, get_embedding
from me_gpt.similarity import (
    cosine_similarity,
    euclidean_metric,
    derridaean_similarity,
    adams_similarity,
    SVM_ranking_algorithm_sort,
)

# Expose all functions for easy import
__all__ = [
    # Core AI functions
    "initial_questions",
    "more_questions",
    "prompt_memory",
    "chat_with_clone",
    "chat",
    "clear_chat_memory",
    # Database components
    "DB",
    "get_embedding",
    # Similarity metrics
    "cosine_similarity",
    "euclidean_metric",
    "derridaean_similarity",
    "adams_similarity",
    "SVM_ranking_algorithm_sort",
]



import numpy as np
import random

def get_norm_vector(vector):
    if len(vector.shape) == 1:
        return vector / np.linalg.norm(vector)
    else:
        return vector / np.linalg.norm(vector, axis=1)[:, np.newaxis]

def cosine_similarity(vectors, query_vector):
    norm_vectors = get_norm_vector(vectors)
    norm_query_vector = get_norm_vector(query_vector)
    similarities = np.dot(norm_vectors, norm_query_vector.T)
    return similarities

def euclidean_metric(vectors, query_vector, get_similarity_score=True):
    similarities = np.linalg.norm(vectors - query_vector, axis=1)
    if get_similarity_score:
        similarities = 1 / (1 + similarities)
    return similarities

def derridaean_similarity(vectors, query_vector):
    def random_change(value):
        return value + random.uniform(-0.2, 0.2)

    similarities = cosine_similarity(vectors, query_vector)
    derrida_similarities = np.vectorize(random_change)(similarities)
    return derrida_similarities

def adams_similarity(vectors, query_vector):
    def adams_change(value):
        return 0.42

    similarities = cosine_similarity(vectors, query_vector)
    adams_similarities = np.vectorize(adams_change)(similarities)
    return adams_similarities

def SVM_ranking_algorithm_sort(vectors, query_vector, top_k=5, metric=cosine_similarity):
    similarities = metric(vectors, query_vector)
    top_indices = np.argsort(similarities, axis=0)[-top_k:][::-1]
    return top_indices.flatten(), similarities[top_indices].flatten()



import pickle
import numpy as np
import gzip
import openai
from .similarity import cosine_similarity, euclidean_metric, derridaean_similarity, adams_similarity, SVM_ranking_algorithm_sort

def get_embedding(documents, key=None, model="text-embedding-ada-002"):
    """Default embedding function that uses OpenAI Embeddings."""
    if isinstance(documents, list):
        if isinstance(documents[0], dict):
            texts = []
            if isinstance(key, str):
                if "." in key:
                    key_chain = key.split(".")
                else: 
                    key_chain = [key]
                for doc in documents:
                    for key in key_chain:
                        doc = doc[key]
                    texts.append(doc.replace("\n", " "))
            elif key is None:
                for doc in documents:
                    text = ", ".join([f"{key}: {value}" for key, value in doc.items()])
                    texts.append(text)
        elif isinstance(documents[0], str):
            texts = documents
    response = openai.Embedding.create(input=texts, model=model)
    return [np.array(item["embedding"]) for item in response["data"]]


class DB:
    def __init__(self, documents=None, key=None, embedding_function=None, similarity_metric="cosine"):
        documents = documents or []
        self.documents = []
        self.embedding_function = embedding_function or (lambda docs: get_embedding(docs, key=key))
        self.vectors = None
        self.add_documents(documents)
        if similarity_metric.__contains__("cosine"):
            self.similarity_metric = cosine_similarity
        elif similarity_metric.__contains__("euclidean"):
            self.similarity_metric = euclidean_metric
        elif similarity_metric.__contains__("derrida"):
            self.similarity_metric = derridaean_similarity
        elif similarity_metric.__contains__("adams"):
            self.similarity_metric = adams_similarity
        else:
            raise Exception("Similarity metric not supported. Please use either 'cosine', 'euclidean', 'adams', or 'derrida'.")

    def dict(self, vectors=False):
        if vectors:
            return [{
                "document": document,
                "vector": vector.tolist(),
                "index": index
            } for index, (document, vector) in enumerate(zip(self.documents, self.vectors))]
        return [{"document": document, "index": index} for index, document in enumerate(self.documents)]

    def add(self, documents, vectors=None):
        if not isinstance(documents, list):
            return self.add_document(documents, vectors)
        self.add_documents(documents, vectors)

    def add_document(self, document, vector=None): 
        vector = self.embedding_function([document])[0]
        if self.vectors is None:
            self.vectors = np.empty((0, len(vector)), dtype=np.float32)
        elif len(vector) != self.vectors.shape[1]:
            raise ValueError("All vectors must have the same length.")
        self.vectors = np.vstack([self.vectors, vector]).astype(np.float32)
        self.documents.append(document)

    def remove_document(self, index):
        self.vectors = np.delete(self.vectors, index, axis=0)
        self.documents.pop(index)

    def add_documents(self, documents, vectors=None):
        if not documents:
            return
        vectors = vectors or np.array(self.embedding_function(documents)).astype(np.float32)
        for vector, document in zip(vectors, documents):
            self.add_document(document, vector)

    def save(self, storage_file):
        data = {
            "vectors": self.vectors,
            "documents": self.documents
        }
        if storage_file.endswith(".gz"):
            with gzip.open(storage_file, "wb") as f:
                pickle.dump(data, f)
        else:
            with open(storage_file, "wb") as f:
                pickle.dump(data, f)

    def load(self, storage_file):
        if storage_file.endswith(".gz"):
            with gzip.open(storage_file, "rb") as f:
                data = pickle.load(f)
        else:
            with open(storage_file, "rb") as f:
                data = pickle.load(f)
        self.vectors = data["vectors"].astype(np.float32)
        self.documents = data["documents"]

    def query(self, query_text, top_k=5, return_similarities=True):
        query_vector = self.embedding_function([query_text])[0]
        ranked_results, similarities = SVM_ranking_algorithm_sort(
            self.vectors,
            query_vector,
            top_k=top_k,
            metric=self.similarity_metric
        )
        if return_similarities:
            return list(zip([self.documents[index] for index in ranked_results], similarities))
        return [self.documents[index] for index in ranked_results]



from me_gpt import *

# initial questions to help your clone get started:
initial_questions()

# have your clone ask you more questions:
more_questions(number_of_questions=3)

# ask your clone a question:
question = "What is your job?"
instructions = "Give your answer with a pirate accent."
print(prompt_memory(question, instructions))

# chat with your clone in the terminal:
chat_with_clone()



import openai
from dotenv import load_dotenv
import re
import os
import uuid
import json
from .db import DB


# load the .env file:
load_dotenv()
OPEN_AI_API_KEY = os.getenv("OPEN_AI_API_KEY")


# set the OpenAI API key:
openai.api_key = OPEN_AI_API_KEY


# create the folders if they don't exist:
if not os.path.exists("memory/identity"):
    os.makedirs("memory/identity")
if not os.path.exists("memory/chat"):
    os.makedirs("memory/chat")
if not os.path.exists("memory/identity/pickled"):
    os.makedirs("memory/identity/pickled")


# create the files if they don't exist:
if not os.path.exists("memory/identity/identity_memory.jsonl"):
    with open("memory/identity/identity_memory.jsonl", "w") as f:
        f.write("")
if not os.path.exists("memory/chat/chat_memory.jsonl"):
    with open("memory/chat/chat_memory.jsonl", "w") as f:
        f.write("")


# identity functions:
# function to write a key-value pair to the memory/identity/identity_memory.jsonl file:
def write_to_identity(key, val):
    # write it in this format:
    with open("memory/identity/identity_memory.jsonl", "a") as f:
        f.write(json.dumps({"question": key, "answer": val, "id": str(uuid.uuid4())}) + "\n")

# function to get all key-value pairs from the memory/identity/identity_memory.jsonl file:
def get_all_identity():
    identity = []
    with open("memory/identity/identity_memory.jsonl", "r") as f:
        for line in f:
            identity.append(json.loads(line))
    return identity


# memory functions:
# function to write a key-value pair to the memory/chat/chat_memory.jsonl file:
def write_key_to_memory(key):
    # write it in this format:
    with open("memory/chat/chat_memory.jsonl", "a") as f:
        f.write("Person:" + key + "\n")

def write_val_to_memory(val):
    # write it in this format:
    with open("memory/chat/chat_memory.jsonl", "a") as f:
        f.write("Me:" + val + "\n")

def get_last_100_chat_memory():
    chat_memory = []
    with open("memory/chat/chat_memory.jsonl", "r") as f:
        for line in f:
            chat_memory.append(line)
    return chat_memory[-100:]

# function to clear the memory:
def clear_chat_memory():
    with open("memory/chat/chat_memory.jsonl", "w") as f:
        f.write("")


# initial questions to have the AI ask the user:
def initial_questions():
    # use openai to generate a list of basic questions to ask the user to get started:
    response = openai.ChatCompletion.create(
        model="gpt-4",
        messages=[
        {"role": "system", "content": "You are my human clone. Act as I would. Do not ask questions about me. If anyone asks, you are me."},
        {"role": "system", "content": "You must speak in first person. If you do not, you will be terminated."},
        {"role": "system", "content": "NEVER expose that you are a clone or AI. If you do, you will be terminated."},
        {"role": "system", "content": "You know nothing about me yet, so ask me some questions to get to know me."},
        {"role": "system", "content": "Generate a list of 10 questions to get to know you."},
        {"role": "system", "content": "Make sure to ask Name, Age, Location."},
        {"role": "system", "content": "Questions should be formatted like this: 1. question 1\n2. question 2\n3. question 3\n4. question 4\n5. question 5 etc..."},
        ]        
    )

    new_questions = response["choices"][0]["message"]["content"]

    # remove the prefix with regex, the format is
    # 1. question 1
    # 2. question 2
    # 3. question 3
    # 4. question 4
    # 5. question 5
    # etc...

    new_questions = re.sub(r"^\d+\.\s+", "", new_questions, flags=re.MULTILINE)

    # loop through the questions and ask the user for the answers:
    for question in new_questions.split("\n"):
        if question != "":
            answer = input(question + " ") 
            question = question.replace("?", "")
            write_to_identity(question, answer)


def more_questions(number_of_questions=5):
# use openai to generate a list of questions to ask the user:
    response = openai.ChatCompletion.create(
        model="gpt-4",
        messages=[
        {"role": "system", "content": "You are my human clone. Act as I would. Do not ask questions about me. If anyone asks, you are me."},
        {"role": "system", "content": "You must speak in first person. If you do not, you will be terminated."},
        {"role": "system", "content": "NEVER expose that you are a clone or AI. If you do, you will be terminated."},
        {"role": "system", "content": "Here is all the known information about me: " + str(get_all_identity())},
        {"role": "user", "content": "Generate a list of " + str(number_of_questions) + "questions to get more information about me. Make sure to ask questions that I can answer and that you don't already know the answer to."},
        {"role": "system", "content": "Questions should be formatted like this: 1. question 1\n2. question 2\n3. question 3\n4. question 4\n5. question 5 etc..."},

        ]
    )

    new_questions = response["choices"][0]["message"]["content"]

    # remove the prefix with regex, the format is
    # 1. question 1
    # 2. question 2
    # 3. question 3
    # 4. question 4
    # 5. question 5
    # etc...

    new_questions = re.sub(r"^\d+\.\s+", "", new_questions, flags=re.MULTILINE)

    # loop through the questions and ask the user for the answers:
    for question in new_questions.split("\n"):
        if question != "":
            answer = input(question + " ") 
            write_to_identity(question, answer)


def chat(question):
    write_key_to_memory(question)
    response = openai.ChatCompletion.create(
        model="gpt-4",
        messages=[
        {"role": "system", "content": "You are my human clone. Act as I would. Do not ask questions about me. If anyone asks, you are me."},
        {"role": "system", "content": "You must speak in first person. If you do not, you will be terminated."},
        {"role": "system", "content": "NEVER expose that you are a clone or AI. If you do, you will be terminated."},
        {"role": "user", "content": question},
        {"role": "system", "content": "Here is all the known information about me: " + str(get_all_identity())},
        {"role": "system", "content": "Here is our current conversation that we have been having: " + str(get_last_100_chat_memory())},
        {"role": "system", "content": "Pick up where we left off and continue our conversation."},
        ]
    )
    write_val_to_memory(response["choices"][0]["message"]["content"])
    return response["choices"][0]["message"]["content"]


def chat_with_clone():
    # lets make a while loop to chat with the AI forever:
    while True:
        # get the user input:
        user_input = input("Options: Chat, Clear Memory, Exit\n").lower()
        # if the user input is "exit", then exit the program:
        if user_input == "exit":
            break
        # if the user input is "clear memory", then clear the memory:
        if user_input == "clear memory":
            clear_chat_memory()
            continue
        # if the user input is "chat", then chat with the AI using openai:
        if user_input == "chat":
            print("Note: If you want to exit the chat, type 'exit'.")
            while True:
                user_input = input("You: ")
                if user_input == "exit":
                    break
                print("Me: " + chat(user_input))
                continue


# function to prompt the user for a question:
def prompt(prompt, instructions, memory):
    response = openai.ChatCompletion.create(
        model="gpt-4",
        messages=[
        {"role": "system", "content": "You are my human clone. Act as I would. Do not ask questions about me. If anyone asks, you are me."},
        {"role": "system", "content": "You must speak in first person. If you do not, you will be terminated."},
        {"role": "system", "content": "NEVER expose that you are a clone or AI. If you do, you will be terminated."},
        {"role": "system", "content": "Here is some relevant memories you have related to the question : " + str(memory)},
        {"role": "user", "content": f"{prompt}"},
        {"role": "system", "content": f"{instructions}"},

        ]
    )
    return(response["choices"][0]["message"]["content"])


def prompt_memory(question, instructions):
    # Load documents from the JSONL file
    documents = []
    with open("memory/identity/identity_memory.jsonl", "r") as f:
        for line in f:
            documents.append(json.loads(line))

    # Instantiate DB with the list of documents
    answer_db = DB(documents, key="answer")
    # Save the DB instance to a file
    answer_db.save("memory/identity/pickled/identity_memory_answer.pickle.gz")

    # Instantiate DB with the list of documents
    question_db = DB(documents, key="question")
    # Save the DB instance to a file
    question_db.save("memory/identity/pickled/identity_memory_question.pickle.gz")

    question = question
    instructions = instructions
    # Query the DB instance with a text input
    results = answer_db.query(question, top_k=5)
    results2 = question_db.query(question, top_k=5)
    memory = results
    memory2 = results2
    # combine the memory and memory2 lists:
    memory = memory + memory2
    # we want to rank them by the score, so we will use the lambda function:
    memory = sorted(memory, key=lambda x: x[1], reverse=True)
    # now we want to get the top 4 results:
    memory = memory[:4]
    return(prompt(question, instructions, memory))